/**********************************************************************/
/* The Minstrel's Song is (c) 1998 by Jonathan Hayward.  See the file */
/* lib/copyrigh.txt for a copyright and license statement.            */
/**********************************************************************/
#include "tms.h"

/* Garbage collection routines.  */

/* Free a list of items, but not the items themselves.  */

void free_ilist(struct ilist *il)
    {
    if (!(il->next == NULL))
	free_ilist(il->next);
    free(il);
    }

/* Free a list of creatures, but not the creatures themselves.  */

void free_clist(struct clist *cl)
    {
    if (cl != NULL)
	{
	if (!(cl->next == NULL))
	    free_clist(cl->next);
	free(cl);
	}
    }

/* Free a list of items, including the items themselves.  */

void free_items(struct ilist *il)
    {
    if (!(il->next == NULL))
	free_items(il->next);
    free_item(il->current);
    free(il);
    }

/* Free a list of creatures, including the creatures themselves.  */

void free_creatures(struct clist *cl)
    {
    if (!(cl->next == NULL))
	free_creatures(cl->next);
    free_creature(cl->current);
    free(cl);
    }

/* Free a list of light sources, but not their holders.  */

void free_light_source(struct light_source_struct *l)
    {
    if (!(l->next == NULL))
	free_light_source(l->next);
    free(l);
    }

/* Free an item and its contents.  */

void free_item(struct small_item *i)
    {
    struct ilist *traverse, *previous, *curr_ilist;
    if (!(i->contents == NULL))
	free_items(i->contents);
    if (!(master_ilist == NULL))
	{
	if (master_ilist->current == i)
	    {
	    curr_ilist = master_ilist;
	    master_ilist = master_ilist->next;
	    free(curr_ilist);
	    }
	else
	    for(traverse = previous = master_ilist; (traverse!=NULL);
	      traverse = traverse->next)
		{
		if (traverse->current == i)
		    {
		    previous->next = traverse->next;
		    free(traverse);
		    traverse = previous;
		    }
		previous = traverse;
		}
	}
    if (i->creatures != NULL)
	{
	free_creatures(i->creatures);
	i->creatures = NULL;
	}
    free(i);
    }

/* Free a creature, its possessions, and its mount, but not any followers,
hostile creatures, or leaders.  */

void free_creature(struct small_creature *c)
    {
    struct clist *traverse, *previous, *curr_clist;
    if (!(c->follow == NULL))
	free_clist(c->follow);
    if (!(c->hostile == NULL))
	free_clist(c->hostile);
    if (!(c->follow == NULL))
	free_clist(c->follow);
    if (!(c->followed_by == NULL))
	free_clist(c->followed_by);
    if (!(c->extended == NULL))
	free_master_creature(c->extended);
    if (!(master_clist == NULL))
	{
	if (master_clist->current == c)
	    {
	    curr_clist = master_clist;
	    master_clist = master_clist->next;
	    free(curr_clist);
	    }
	else
	    {
	    for (previous = traverse = master_clist; (!(traverse == NULL));
	      traverse = traverse->next)
		if (traverse->current == c)
		    {
		    previous->next = traverse->next;
		    free(traverse);
		    traverse = previous;
		    }
		else
		    previous = traverse;
	    }
	}
    free(c);
    }

void free_faith_struct(struct faith_struct *f)
    {
    if (f->breath_prayer != NULL)
	free_text_struct(f->breath_prayer);
    free(f);
    }

void free_master_creature(struct master_creature *c)
    {
    if (!(c->mount == NULL))
	free_creature(c->mount);
    if (!(c->wear_wield == NULL))
	free_items(c->wear_wield);
    if (!(c->pack == NULL))
	free_items(c->pack);
    if (!(c->stati == NULL))
	free(c->stati);
    if (!(c->skills == NULL))
	free(c->skills);
    if (c->faith != NULL)
	free_faith_struct(c->faith);
    free(c);
    }

void free_shop(struct shop_struct *s)
    {
    if (!(s->items==NULL))
	free_items(s->items);
    if (!(s->creatures==NULL))
	free_creatures(s->creatures);
    free(s);
    }

/* We do not free the level associated with a special.  */

void free_special(struct special_struct *s)
    {
    if (!(s->shop==NULL))
	free_shop(s->shop);
    free(s);
    }

void free_text_struct(struct text_struct *t)
    {
    if (!(t->next == NULL))
	free_text_struct(t->next);
    free(t);
    }

void free_level(struct level *l)
    {
    long i, j;
    struct llist *this_llist, *traverse, *previous;
    if (!(l->light_sources==NULL))
	free_light_source(l->light_sources);
    if (!(l->creatures==NULL))
	free_clist(l->creatures);
    for(i=0;i<MAX_X;++i)
	for(j=0;j<MAX_Y;++j)
	    {
	    if (!(l->space[i][j].creatures==NULL))
		free_creatures(l->space[i][j].creatures);
	    if (!(l->space[i][j].items==NULL))
		free_items(l->space[i][j].items);
	    if (!(l->space[i][j].special==NULL))
		free_special(l->space[i][j].special);
	    }
    if (!(master_llist == NULL))
	{
	if (master_llist->current == l)
	    {
	    this_llist = master_llist;
	    master_llist = master_llist->next;
	    free(this_llist);
	    }
	else
	    for(traverse = previous = master_llist; (!(traverse == NULL));
	      traverse = traverse->next)
		if (traverse->current == l)
		    {
		    previous->next = traverse->next;
		    free(traverse);
		    traverse = previous;
		    }
		else
		    previous = traverse;
	}
    free(l);
    }

/* Free all of the linked list, given its head.  */

void free_read_file_text_struct(struct read_file_text_struct *r)
    {
    if (r != NULL)
	{
	if (r->next != NULL)
	    free_read_file_text_struct(r->next);
	free(r);
	}
    }

/* Free the pointer structure used to regenerate data from a save file, but */
/* not the new data themselves.  */

void free_pointer_struct(struct pointer_struct *pointer)
    {
    if (!(pointer->next == NULL))
	free_pointer_struct(pointer->next);
    free(pointer);
    }
